/*----------------------------------------------------------------------
  File    : bridgit.c
  Contents: bridg-it game with computer player
  Author  : Christian Borgelt
  History : 2001.05.29 file created
            2001.06.03 first version completed
            2013.03.07 resistor sorting made compatible with 64bit
            2013.08.09 adapted to higher compiler warning level
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include "arrays.h"
#include "bridgit.h"

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define EPSILON  1e-12         /* to handle roundoff errors */

/* compute resistor index from row and column coordinates */
#define bi_index(b,c,r) (((c) & 1) \
                        ? (((r)/2)   * (b)->size    +((c)/2))  \
                        : (((r)/2-1) *((b)->size-1) +((c)/2-1) \
                           + (b)->size * (b)->size))

/* check whether point given by row and column is a proper crossing */
#define bi_cross(b,c,r) (  ((c) > 0) && ((c) <= 2 *(b)->size) \
                        && ((r) > 0) && ((r) <= 2 *(b)->size) \
                        && !(((r) + (c)) & 1))

/*----------------------------------------------------------------------
  Auxiliary Functions
----------------------------------------------------------------------*/

static int ccmp (int i1, int i2, void *data)
{                              /* --- comparison of currents */
  double c1, c2;               /* currents to compare */

  c1 = ((const double*)data)[(long)i1];
  c2 = ((const double*)data)[(long)i2];
  if     (c1 > c2 +EPSILON) return 1; /* return the sign of the diff. */
  return (c1 < c2 -EPSILON) ? -1 : 0; /* of the two currents */
}  /* ccmp() */

/*----------------------------------------------------------------------
  Main Functions
----------------------------------------------------------------------*/

BRIDGIT* bi_create (int size)
{                               /* --- create a bridg-it game */
  int     n, nl, nc;            /* numbers of resistors */
  BRIDGIT *bi;                  /* created bridg-it game */

  assert((size >= BI_MINSIZE)   /* check size against minimum */
      && (size <= BI_MAXSIZE)); /* and maximum board size */
  nl =  size    * size;         /* number of longitudinal resistors */
  nc = (size-1) *(size-1);      /* number of crossway     resistors */
  n  = nl +nc;                  /* total number of resistors */
  bi = (BRIDGIT*)malloc(sizeof(BRIDGIT) +(size_t)(3*n-1) *sizeof(int));
  if (!bi) return NULL;         /* create the base structure */
  bi->size   = size;            /* and note the number of points */
  bi->ptcnt  = n;
  bi->states = bi->moves  +n;   /* organize the different vectors */
  bi->buf    = bi->states +n;
  bi->vert   = mat_create(n,n);
  bi->horz   = mat_create(n,n); /* create matrices and vectors */
  bi->tmp    = mat_create(n,n); /* for the current equations */
  bi->vec    = (double*)malloc((size_t)n *sizeof(double));
  bi->sol    = (double*)malloc((size_t)n *sizeof(double));
  if (!bi->horz || !bi->vert || !bi->tmp || !bi->vec || !bi->sol) {
    bi_delete(bi); return NULL; }
  return bi;                    /* return the created bridg-it game */
}  /* bi_create() */

/*--------------------------------------------------------------------*/

void bi_delete (BRIDGIT *bi)
{                               /* --- delete a bridg-it game */
  assert(bi);                   /* check the function arguments */
  if (bi->sol)  free(bi->sol);
  if (bi->vec)  free(bi->vec);
  if (bi->tmp)  mat_delete(bi->tmp);
  if (bi->vert) mat_delete(bi->vert);
  if (bi->horz) mat_delete(bi->horz);
  free(bi);                     /* delete the matrix, the vectors, */
}  /* bi_delete() */            /* and the base structure */

/*--------------------------------------------------------------------*/

void bi_init (BRIDGIT *bi)
{                               /* --- initialize a bridg-it game */
  int x, y, r = 0;              /* loop variables, index */
  int size;                     /* number of points on the panels */
  int nl, nc;                   /* number of long./cross. resistors */
  int ol, oc;                   /* offsets for resistor indices */

  assert(bi);                   /* check the function arguments */
  size = bi->size;              /* get the number of points and the */
  nl   =  size    * size;       /* numbers of long./cross. resistors */
  nc   = (size-1) *(size-1);    /* init. the matrices and the vector */
  vec_init(bi->vec, nl +nc, nl +nc -1);
  mat_init(bi->horz, MAT_ZERO, NULL);
  mat_init(bi->vert, MAT_ZERO, NULL);

  /* --- horizontal mesh equations / vertical node equations --- */
  ol = nl; oc = nl +nc;         /* traverse the network rows */
  for (y = size-1; --y >= 0; ){ /* and the columns of each row */
    ol -= size; oc -= size-1;   /* and set the coefficients */
    for (x = size; --x >= 0; r++) {
      mat_set(bi->horz, r, ol +x -size, -1);
      mat_set(bi->vert, r, ol +x -size,  1);
      mat_set(bi->horz, r, ol +x,        1);
      mat_set(bi->vert, r, ol +x,       -1);
      if (x > 0)      { mat_set(bi->horz, r, oc +x-1,  1);
                        mat_set(bi->vert, r, oc +x-1,  1); }
      if (x < size-1) { mat_set(bi->horz, r, oc +x,   -1);
                        mat_set(bi->vert, r, oc +x,   -1); }
    }                           /* (do not set fourth coefficient */
  }                             /* for first/last network column) */

  /* --- horizontal node equations / vertical mesh equations --- */
  ol = nl; oc = nl +nc;         /* traverse the network rows */
  for (y = size; --y >= 0; ) {  /* and the columns of each row */
    ol -= size; oc -= size-1;   /* and set the coefficients */
    for (x = size-1; --x >= 0; r++) {
      mat_set(bi->horz, r, ol +x,     1);
      mat_set(bi->vert, r, ol +x,     1);
      mat_set(bi->horz, r, ol +x +1, -1);
      mat_set(bi->vert, r, ol +x +1, -1);
      if (y > 0)      { mat_set(bi->horz, r, oc +x,          1);
                        mat_set(bi->vert, r, oc +x,         -1); }
      if (y < size-1) { mat_set(bi->horz, r, oc +x +size-1, -1);
                        mat_set(bi->vert, r, oc +x +size-1,  1); }
    }                           /* (do not set fourth coefficient */
  }                             /* for first/last network row) */

  /* --- boundary equations --- */
  for (ol = nl -size, y = size; --y >= 0; ol -= size)
    mat_set(bi->horz, r, ol, 1);/* horizontal network */
  for (x = size; --x >= 0; )
    mat_set(bi->vert, r, x,  1);/* vertical   network */

  /* --- other initializations --- */
  bi->mvcnt = bi->curr = 0;     /* clear the number of moves */
  for (x = nl +nc; --x >= 0; )  /* and the state vector */
    bi->states[x] = BI_NONE;    /* (no bridges built yet) */
}  /* bi_init() */

/*----------------------------------------------------------------------
The resistors (or, what is the same, the crossing points) are numbered
from left to right in each row, and from top to bottom over the rows.
For the horizontal network, the horizontal resistors are numbered first,
the vertical resistors after all horizontal ones have been numbered.
For the vertical network it is the other way round (vertical resistors
are numbered before horizontal ones). In this way resistors that
correspond to the same crossing point have the same numbers.
The currents always flow from left to right and from top to bottom.
----------------------------------------------------------------------*/

int bi_hint (BRIDGIT *bi, int player, RANDFN randfn)
{                               /* --- give a hint for a player */
  int    i, k;                  /* loop variables */
  MATRIX *mat;                  /* matrix of the system to solve */
  int    col, row;              /* row and column of bridge */

  assert(bi && randfn           /* check the function arguments */
      && ((player == BI_HORZ) || (player == BI_VERT)));
  mat = (player & BI_HORZ)      /* get the corresponding matrix */
      ? bi->horz : bi->vert;    /* and solve the equation system */
  #if 0
  mat_gjsol(&bi->sol, mat, &bi->vec, 1, MAT_FULLPIV, NULL);
  #endif                        /* solve directly with Gauss alg. */
  #if 0
  mat_gjinv(bi->tmp, mat, MAT_FULLPIV, NULL);
  mat_mulmv(bi->sol, bi->tmp, bi->vec);
  #endif                        /* solve by multiplying with inverse */
  #if 1
  mat_ludecom(bi->tmp, mat);    /* decompose the coefficient matrix */
  mat_lusubst(bi->sol, bi->tmp, bi->vec);
  #endif                        /* do forward/backward substitution */
  for (i = bi->ptcnt; --i >= 0; ) {
    bi->sol[i] = fabs(bi->sol[i]);
    bi->buf[i] = i;             /* make all currents positive */
  }                             /* and initialize the index vector */
  for (i = bi->curr; --i >= 0; ) {
    k = bi->moves[i];           /* traverse the moves of the opponent */
    if (k & player) continue;   /* (i.e., the blocked connections) */
    k = bi_index(bi, bi_col(k), bi_row(k));
    if (bi->sol[k] > 1e-5)      /* a non-negligible current through */
      return BI_NONE;           /* a blocked connection indicates */
  }                             /* that the game has been lost */
  int_shuffle(bi->buf, (size_t)bi->ptcnt, randfn);
  i2c_qsort(bi->buf, (size_t)bi->ptcnt, +1,
            ccmp, bi->sol);     /* shuffle and sort the resistors */
  for (k = bi->ptcnt; --k >= 0; ) {
    if (bi->states[bi->buf[k]] == BI_NONE)
      break;                    /* traverse the resistors/crossings */
  }                             /* and find one without a bridge */
  if (k < 0) return BI_NONE;    /* check whether one was found and */
  k = bi->buf[k];               /* get the resistor/crossing index */
  i = bi->size *bi->size;       /* get number of long. resistors */
  if (k < i) {                  /* if longitudinal resistor */
    col =  (k    %  bi->size)    *2 +1;
    row =  (k    /  bi->size)    *2 +1; }
  else {                        /* if crossway resistor */
    col = ((k-i) % (bi->size-1)) *2 +2;
    row = ((k-i) / (bi->size-1)) *2 +2;
  }                             /* get coordinates of crossing point */
  return bi_bridge(player & (BI_VERT|BI_HORZ), col, row);
}  /* bi_hint() */              /* return the corresponding bridge */

/*--------------------------------------------------------------------*/

void bi_build (BRIDGIT *bi, int bridge)
{                               /* --- build a bridge */
  int    i, n, r;               /* loop variables, resistor index */
  int    col, row;              /* coordinates of crossing point */
  double h, v, o;               /* new and old resistor values */

  assert(bi);                   /* check the function arguments */
  col = bi_col(bridge);         /* get the coordinates */
  row = bi_row(bridge);         /* of the crossing point */
  assert(bi_cross(bi,col,row)); /* check for a correct crossing point */
  if ((bi->mvcnt <= bi->curr)   /* if the bridge is not rebuilt */
  ||  (bi->moves[bi->curr] != bridge)) {
    bi->moves[bi->curr] = bridge;
    bi->mvcnt = bi->curr +1;    /* note the new bridge */
  }                             /* and adapt the move counter */
  bi->curr++;                   /* increment the move counter */
  r = bi_index(bi, col, row);   /* get the resistor index and */
  bi->states[r] = bridge;       /* set the state of the point */
  if (bi_player(bridge) & BI_HORZ) { v = 1e+10; h = 1e-10; }
  else                             { v = 1e-10; h = 1e+10; }
  n = bi->ptcnt >> 1;           /* get the new resistor values and */
  for (i = n; --i >= 0; ) {     /* traverse the mesh equations */
    o = mat_get(bi->horz, i,   r); /* adapt horizontal network */
    if (o != 0) mat_set(bi->horz, i,   r, (o > 0) ? h : -h);
    o = mat_get(bi->vert, i+n, r); /* adapt vertical   network */
    if (o != 0) mat_set(bi->vert, i+n, r, (o > 0) ? v : -v);
  }                             /* (open/close connections) */
}  /* bi_build() */

/*--------------------------------------------------------------------*/

int bi_state (BRIDGIT *bi, int col, int row)
{                               /* --- get the state of a crossing */
  assert(bi);                   /* check the function arguments */
  assert(bi_cross(bi,col,row)); /* check for a correct crossing point */
  return bi->states[bi_index(bi, col, row)];
}  /* bi_state() */             /* return the state of the crossing */

/*--------------------------------------------------------------------*/

int bi_undo (BRIDGIT *bi)
{                               /* --- undo a move */
  int    bridge;                /* the bridge to remove */
  int    i, n, r;               /* loop variables, resistor index */
  int    col, row;              /* coordinates of crossing point */
  double o;                     /* old resistor value */

  assert(bi);                   /* check the function arguments */
  if (bi->curr <= 0)            /* if there are no more moves, */
    return BI_NONE;             /* abort the function */
  bridge = bi_move(bi, --bi->curr);
  col    = bi_col(bridge);      /* get the bridge to remove and the */
  row    = bi_row(bridge);      /* coordinates of the crossing point */
  assert(bi_cross(bi,col,row)); /* check for a correct crossing point */
  r = bi_index(bi, col, row);   /* get the resistor index and */
  bi->states[r] = BI_NONE;      /* reset the state of the point */
  n = bi->ptcnt >> 1;           /* get the number of meshes and */
  for (i = n; --i >= 0; ) {     /* traverse the mesh equations */
    o = mat_get(bi->horz, i,   r); /* adapt horizontal network */
    if (o != 0) mat_set(bi->horz, i,   r, (o > 0) ? 1 : -1);
    o = mat_get(bi->vert, i+n, r); /* adapt vertical   network */
    if (o != 0) mat_set(bi->vert, i+n, r, (o > 0) ? 1 : -1);
  }                             /* (reset connections) */
  return bridge;                /* return the removed bridge */
}  /* bi_undo() */

/*--------------------------------------------------------------------*/

int bi_redo (BRIDGIT *bi)
{                               /* --- redo a move */
  int bridge;                   /* the bridge to rebuild */

  assert(bi);                   /* check the function arguments */
  if (bi->curr >= bi->mvcnt)    /* if there are no more moves, */
    return BI_NONE;             /* abort the function */
  bridge = bi_move(bi,bi->curr);/* get the corresponding bridge */
  bi_build(bi, bridge);         /* and rebuild it */
  return bridge;                /* return the rebuilt bridge */
}  /* bi_redo() */

/*--------------------------------------------------------------------*/

int bi_save (BRIDGIT *bi, FILE *file)
{                               /* --- save a bridg-it game */
  int i;                        /* loop variable */

  fprintf(file, "%d %d\n", bi->size, bi->curr);
  for (i = 0; i < bi->mvcnt; i++)
    fprintf(file, "%06x\n", bi->moves[i]);
  return ferror(file) ? -1 : 0; /* print all moves made and */
}  /* bi_save() */              /* return the write status */

/*--------------------------------------------------------------------*/

BRIDGIT* bi_load (FILE *file)
{                               /* --- load a bridg-it game */
  int     size, cnt;            /* board size, move counter */
  unsigned int bridge;          /* to traverse the bridges */
  int     player;               /* player that built the bridge */
  int     row, col;             /* row and column of crossing point */
  BRIDGIT *bi;                  /* loaded game */

  if ((fscanf(file, "%d%d", &size, &cnt) != 2)
  ||  (size < 2)) return NULL;  /* read and check the board size */
  bi = bi_create(size);         /* and the current move number */
  if (!bi) return NULL;         /* and create a bridg-it game */
  bi_init(bi);                  /* initialize the bridg-it game */
  while (bi->mvcnt < bi->ptcnt){/* traverse the possible moves */
    if (fscanf(file, "%x", &bridge) != 1) {
      if (ferror(file)) { bi_delete(bi); return NULL; }
      break;                    /* read the next bridge and */
    }                           /* check for error and end of file */
    player = bridge & 0xff;     /* get the parameters of the bridge */
    col    = bi_col(bridge);    /* (player that built the bridge */
    row    = bi_row(bridge);    /* and coords. of the crossing) */
    if (((player != BI_HORZ) && (player != BI_VERT))
    ||  !bi_cross(bi, col, row) /* check for a valid player and */
    ||  (bridge >> 24 != 0)) {  /* for a valid crossing point */
      bi_delete(bi); return NULL; }
    bi->moves[bi->mvcnt++] = (int)bridge;
  }                             /* store the loaded bridge */
  if (cnt > bi->mvcnt) {        /* check the current move number */
    bi_delete(bi); return NULL; }
  while (--cnt >= 0)            /* traverse the loaded moves */
    bi_redo(bi);                /* and execute them */
  return bi;                    /* return the loaded game */
}  /* bi_load() */

/*--------------------------------------------------------------------*/
#ifndef NDEBUG

void bi_show (BRIDGIT *bi, int player)
{                               /* --- show currents in network */
  int x, y, o;                  /* loop variables, offset */
  int nl;                       /* number of longitudinal resistors */
  int s;                        /* state of crossing point */

  nl = bi->size * bi->size;     /* get the number of long. resistors */
  if (player == BI_HORZ) {      /* if to show the horizontal network */
    for (y = 0; y < bi->size-1; y++) {
      for (o = y *bi->size, x = 0; x < bi->size; x++)
        printf("*-%5.3f-", fabs(bi->sol[o+x]));
      printf("*\n ");           /* horizontal currents */
      for (x = bi->size-1; --x >= 0; ) printf("       |");
      printf("\n   ");          /* vertical connections */
      for (o = nl +y *(bi->size-1), x = 0; x < bi->size-1; x++)
        printf("   %5.3f", fabs(bi->sol[o+x]));
      printf("\n ");            /* vertical currents */
      for (x = bi->size-1; --x >= 0; ) printf("       |");
      printf("\n");             /* vertical connections */
    }
    for (o = y *bi->size, x = 0; x < bi->size; x++)
      printf("*-%5.3f-", fabs(bi->sol[o+x]));
    printf("*\n"); }            /* horizontal currents in last row */
  else if (player == BI_VERT) { /* if to show the vertical network */
    for (x = bi->size-1; --x >= 0; ) printf("    *   ");
    printf("    *\n    ");      /* first line (connection panel) */
    for (y = 0; y < bi->size; y++) {
      for (x = bi->size-1; --x >= 0; ) printf("|       ");
      printf("|\n  ");          /* vertical connections */
      for (o = y *bi->size, x = 0; x < bi->size-1; x++)
        printf("%5.3f   ",  fabs(bi->sol[o+x]));
      printf("%5.3f\n    ", fabs(bi->sol[o+x]));
      for (x = bi->size-1; --x >= 0; ) printf("|       ");
      printf("|\n    ");        /* vertical connections */
      if (y >= bi->size-1) break;
      for (o = nl +y *(bi->size-1), x = 0; x < bi->size-1; x++)
        printf("*-%5.3f-", fabs(bi->sol[o+x]));
      printf("*\n    ");        /* horizontal currents */
    }
    for (x = bi->size-1; --x >= 0; ) printf("*       ");
    printf("*\n"); }          /* last line (connection panel) */
  else {                        /* if to show the bridges built */
    printf("    ");             /* (that is, the current board) */
    for (x = bi->size-1; --x >= 0; ) printf("oooooooo");
    printf("o\n");              /* print connection panel */
    for (x = 0; x < bi->size; x++) {
      if (bi->states[x] & BI_VERT) printf("    o   ");
      else                         printf("        ");
    }                           /* first part of vertical bridges */
    printf("\n");               /* of vertical network (if present) */
    for (y = 0; y < bi->size; y++) {
      for (o = x = 0; x < bi->size; x++) {
        s = bi->states[y *bi->size +x];
        if      (s & BI_HORZ) printf("********");
        else if (s & BI_VERT) printf("*   o   ");
        else                  printf("*       ");
      }                         /* horz. bridges of horz. network */
      printf("*\n");            /* and 2nd part of vertical bridges */
      if (y >= bi->size-1) break;
      printf("*");              /* abort loop after last row */
      for (x = 0; x < bi->size; x++) {
        s = bi->states[y *bi->size +x];
        if (s & BI_VERT) printf("   o");
        else             printf("    ");
        if (x >= bi->size-1)    /* third part of vertical bridges */
          break;                /* of vertical network (if present) */
        s = bi->states[nl +y *(bi->size-1) +x];
        if (s & BI_HORZ) printf("   *");
        else             printf("    ");
      }                         /* first part of vertical bridges */
      printf("   *\n*   ");     /* of horizontal network (if present) */
      for (x = 0; x < bi->size-1; x++) {
        s = bi->states[nl +y *(bi->size-1) +x];
        if      (s & BI_VERT) printf("oooooooo");
        else if (s & BI_HORZ) printf("o   *   ");
        else                  printf("o       ");
      }                         /* horz. bridges of vertical network */
      printf("o   *\n*");       /* and 2nd part of vertical bridges */
      for (x = 0; x < bi->size; x++) {
        s = bi->states[(y+1) *bi->size +x];
        if (s & BI_VERT) printf("   o");
        else             printf("    ");
        if (x >= bi->size-1)    /* first part of vertical bridges */
          break;                /* of vertical network (if present) */
        s = bi->states[nl +y *(bi->size-1) +x];
        if (s & BI_HORZ) printf("   *");
        else             printf("    ");
      }                         /* third part of vertical bridges */
      printf("   *\n");         /* of horizontal network (if present) */
    }
    for (x = 0; x < bi->size; x++) {
      s = bi->states[y *bi->size +x];
      if (s & BI_VERT) printf("    o   ");
      else             printf("        ");
    }                           /* third part of vertical bridges */
    printf("\n    ");           /* of vertical network (if present) */
    for (x = bi->size-1; --x >= 0; ) printf("oooooooo");
    printf("o\n");              /* print connection panel */
  }
}  /* bi_show() */

#endif
